#pragma off (unreferenced)
static char iogsadad_c_id[] = "$Id: iogsadad.c,v 1.2 2002/01/17 13:26:46 sgn Exp sgn $";
#pragma on (unreferenced)
/* $Locker: sgn $
 * 
 *      This program is Copyright (C) 2001 COM, Incorporated
 *                    ALL RIGHTS RESERVED
 *
 * CAUTION: THIS FILE CONTAINS HARDWARE-, OS-, AND/OR
 *          COMPILER-SPECIFIC MANIFESTS, TABLES, AND CODE!!
 *
 * This file contains functions which are specific to the
 * General Standards Corporation PCI-ADADIO analog I/O 
 * cards.
 *
 * MODIFICATION HISTORY:
 * $Log: iogsadad.c,v $
 * Revision 1.2  2002/01/17 13:26:46  sgn
 * Added some debug code and rearranged the register accesses.
 *
 * Revision 1.1  2002/01/07 14:05:42  sgn
 * Initial revision
 *
 *
 */

// Activate global declarations as external here (see: EXTERN in *.h)
#define EXTERN extern

// Activate register definitions
#define INCLUDE_REGDEFS

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <conio.h>
#include <sys/kernel.h>
#include <sys/pci.h>
#include <string.h>
#include <stddef.h>
#include <sys/seginfo.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

#include "basetyps.h"
#include "tscdefs.h"
#include "iogenerl.h"
#include "iogsadad.h"
#include "ioshmem.h"
#include "ioimmem.h"
#include "ioglobal.h"
#include "iodacmem.h"
#include "ioadcmem.h"
#include "iomgr.h"
#include "ioutil.h"

//*****************************************************************************
// Local storage
//*****************************************************************************
// uBusNum and uDevFuncNum only need to be used in iogsadIdent() but declaring
// these in the function causes compilation warnings.  I believe this is
// because this module is compiled with the -zu switch which I guess causes all
// pointers to be creaated as FAR pointer and these variables are used in a
// QNX API function that apparently uses near pointers.
static unsigned uBusNum;		// PCI bus number for found device
static unsigned uDevFuncNum;		// PCI device or function ID for found



//*****************************************************************************
// PCI-ADADIO - IDENTIFY BOARDS
//	This function finds all of the PCI-ADADIO analog I/O cards and
//	creates an array of structures containing data about each card
//	found.
// Args:
//	VerboseMode	This function always prints error messages.  This
//			function will also print more informational/debug
//			messages when this is set higher than V0_NONE.
//	*psCards	This is a pointer to a pointer to the first element of
//			an array of data structures.  Each array element is a
//			data structure that corresponds to a analog I/O
//			card.  The caller need not manipulate this
//			structure.  This function fills fields in this
//			structure for use by other functions in this module.
// Returns:	
//	Number of General Standards PCI-ADADIO cards actually found
//*****************************************************************************
int iogsadadIdent(VERBOSE_MODE VerboseMode, AIOCARD **psCards) 
{
	int iCardNum;			// I/O card list index
	int iReturnCode;		// Function return codes
	AIOCARD *sNewCard;		// Pointer to an I/O card structure.
	unsigned uPCIIndex;		// PCI bus index (used to find multiple
					// cards of the same type)
	unsigned uVendorID = VENDOR_ID;	// Brand specific PCI bus vendor ID
					// code
	unsigned uDeviceID = DEVICE_ID;	// Brand specific PCI bus device ID
					// code
	unsigned uSubsysID;		// Brand specific PCI bus subsystem ID
					// code
	word32 uPCRValue;		// 32 bit PCI control register values
	AIOCARD *sCards = NULL;		// Local pointer to the I/O cards struct.

	// Make sure memory is not already been allocated.
	if (*psCards != NULL)
	{
		fprintf(stderr, "%s: Non-NULL pointer to I/O Cards data"
		"array passed to iogsadadIdent().\n", myname);
		return(0);
	}

	// Look for PCI-ADADIO cards on the bus.

	iCardNum = -1;
	uPCIIndex = 0;

	while(1)	
	{
		// Try to find the next device.
		iReturnCode = _CA_PCI_Find_Device(uDeviceID, uVendorID, 
		uPCIIndex, &uBusNum, &uDevFuncNum);
		if (iReturnCode == PCI_DEVICE_NOT_FOUND) 
		{ // No PCI-ADADIO card found.
			break;
		}
		else if (iReturnCode != PCI_SUCCESS)
		{ // Call error of some sort.
			fprintf(stderr, "%s: Error locating PCI-ADADIO I/O "
			"cards.\n", myname);
			*psCards = sCards;
			return(iCardNum+1);
		}
		else		
		{ // Else a card was found.
			// Verify that the Subsystem ID matches that of a
			// PCI-ADADIO card.
			uSubsysID = 0;
			iReturnCode = _CA_PCI_Read_Config_Word(uBusNum, 
			uDevFuncNum, 0x2e, 1, (char *)&uSubsysID);
			if (iReturnCode != PCI_SUCCESS)
			{
				fprintf(stderr, "%s: Error reading subsystem "
				"ID from a PCI control register 0x2e via "
				"_CA_PCI_Read_Config_Word().\n", myname);
				*psCards = sCards;
				return(iCardNum+1);
			}
			if (uSubsysID != SUBSYS_ID)
			{
				// Apparently this is not a PCI-ADADIO card.
				uPCIIndex++;
				continue;
			}

			// Found a card on the PCI bus.
			iCardNum++;
			if (VerboseMode > V0_NONE)
			{
				fprintf(stdout, "%s: General Standards "
				"PCI-ADADIO AIO card # %d successfully "
				"identified\n", myname, iCardNum+1);
			}

			// Adjust the array size of the list of cards found.
			sNewCard = (AIOCARD *)realloc(sCards, sizeof(AIOCARD)
			* (iCardNum + 1));
			if (sNewCard == NULL)
			{
				fprintf(stderr, "%s: Error allocating memory "
				"for a PCI-ADADIO I/O card.\n", myname);
				*psCards = sCards;
				return(iCardNum+1);
			}
			else
			{
				sCards = sNewCard;
			} 
			// Now find it's base address in memory.  Offset 18h
			// of the PCI control register contains the base I/O
			// address of the card in bits 31:2 (along with
			// other information which can be masked off).
			uPCRValue = 0;	
			iReturnCode = _CA_PCI_Read_Config_DWord(uBusNum, 
			uDevFuncNum, 0x18, 1, (char *)&uPCRValue);
			if (iReturnCode != PCI_SUCCESS)
			{
				fprintf(stderr, "%s: Error reading base I/O "
				"address for a analol I/O card from PCI "
				"control register 0x18 via "
				"_CA_PCI_Read_Config_DWord().\n", myname);
				*psCards = sCards;
				return(iCardNum+1);
			}
			sCards[iCardNum].ulPhysicalAddress = uPCRValue &
			0xfffffffc;
			
			// Look up the card's interrupt number.
			uPCRValue = 0;	
			iReturnCode = _CA_PCI_Read_Config_Byte(uBusNum, 
			uDevFuncNum, 0x3c, 1, (char *)&uPCRValue);
			if (iReturnCode != PCI_SUCCESS)
			{
				fprintf(stderr, "%s: Error reading interrupt "
				"number for a analog I/O card from PCI "
				"control register 0x3c via "
				"_CA_PCI_Read_Config_Byte().\n", myname);
				*psCards = sCards;
				return(iCardNum+1);
			}
			sCards[iCardNum].iInterruptNumber = uPCRValue;

			// Look up the card's PCI Local Configuration
			// Registers (LCR) base address.  Refer to the
			// PLX-9080 PCI interface data sheet for more
			// information.
			uPCRValue = 0;	
			iReturnCode = _CA_PCI_Read_Config_DWord(uBusNum, 
			uDevFuncNum, 0x14, 1, (char *)&uPCRValue);
			if (iReturnCode != PCI_SUCCESS)
			{
				fprintf(stderr, "%s: Error reading base I/O"
				"address for a analog I/O card PCI Local "
				"Configuration Register from PCI control "
				"register 0x14 via "
				"_CA_PCI_Read_Config_DWord().\n", myname);
				*psCards = sCards;
				return(iCardNum+1);
			}
			sCards[iCardNum].iLCRBase = uPCRValue & 0xfffc;

		} // End else PCI-ADADIO device found
		uPCIIndex++;
	} // End while

	*psCards = sCards;
	return(iCardNum+1);
}

//*****************************************************************************
// PCI-ADADIO - I/O INITIALIZATION
//	This function initializes all of the PCI-ADADIO analog I/O cards.
// Args:
//	iCardCnt	Number of analog I/O cards installed.
//	sCards		This is a pointer to the first element of an
//			array of data structures.  Each array element is a
//			data structure that corresponds to a analog I/O
//			card.  The caller need not manipulate this
//			structure.
// Returns:	
//	0 successfully initialized cards
//	1 error initializing cards.	
//*****************************************************************************
int iogsadadInit(int iCardCnt, AIOCARD *sCards)
{
	int iCardNum; // I/O card array index.
	int iFileDesc; // File descriptor for physical shared memory.
	void *pMappedPage; // Pointer to mapped page(s) containing registers. 
	unsigned long ulOffsetInPage; // Offset within page to registers.
	size_t tMemSize; // Size of mapped memory region.
	off_t tPageAddress; // Address of page to be mapped.

	// Make sure memory has been allocated.
	if (sCards == NULL)
	{
		fprintf(stderr, "%s: NULL pointer to I/O Cards data"
		"array passed to iogsadadInit().\n", myname);
		return(1);
	}

	// open physical memory
	iFileDesc = shm_open ( "Physical", O_RDWR, 0777);
	if ( iFileDesc == -1)
 	{
		fprintf(stderr, "%s: error opening physical memory to map"
		"PCI-ADADIO configuration registers", myname, strerror(errno));
		return(1);
	}

	// For each PCI-ADADIO card found...
	for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++)
	{
		
		// Map the register set of each card into this process space.
		// Since mmap() maps a minimum of a page when mapping physical
		// memory, first it must be determined if the 128 byte register
		// space happens to cross a page boundry.  If the space crosses
		// a page boundry then two pages must be mapped.  In this
		// case this is done by adding 128 to the offset within the
		// page and checking to see if the result is greater than
		// 0x1000 (page size).  If the result is greater, map in two
		// pages, otherwise only one page is required to map in the
		// register set.
        	tPageAddress = sCards[iCardNum].ulPhysicalAddress &
		0xfffff000;
		ulOffsetInPage = sCards[iCardNum].ulPhysicalAddress &
		0xfff;

		tMemSize = 0x1000;
		if ((ulOffsetInPage + 128) >= 0x1000)
		{
			tMemSize = 0x2000;
		}

        	pMappedPage = mmap (0, tMemSize, PROT_READ | PROT_WRITE |
		PROT_NOCACHE, MAP_SHARED, iFileDesc, tPageAddress);
        	if (pMappedPage == (void *) -1)
        	{
			fprintf(stderr, "%s: Error mapping PCI-ADADIO "
			"registers via mmap(): %s\n", myname, strerror(errno));
        	        return(1);
        	}
		sCards[iCardNum].pMappedPage = pMappedPage;
		
		sCards[iCardNum].pRegisters = (unsigned long *)
		((char *)pMappedPage + ulOffsetInPage);
		sCards[iCardNum].tMemSize = tMemSize;

		// Initialize each card.  Select the following modes:
		// AIMx (analog input mode): mode 1 - Single ended, burst mode 
		// LBCx (loopback channel): 0
		// CAL STATUS FLAG: Cleared
		// OFFSET BINARY: True
		// SIZEx (input buffer size): 8 samples
		// (Reserved): 0
		// LASTx (last input number): 7 (all inputs used)
		// ENABLE OUTPUTS: True
		// ENABLE OUTPUT STROBE: True
		// INPUT BUFFER EMPTY (read only)
		// INPUT BUFFER HALF FULL (read only)
		// INPUT BUFFER FULL (read only)
		// INTERRUPT Ax (interrupt source): disabled
		// INTERRUPT REQUEST FLAG: clear
		// CMx (calibration mode): none
		// OUTPUT STROBE: 0
		// INPUT TRIGGER: 0
		// INITIALIZE: 0
		sCards[iCardNum].pRegisters[BOARD_CONTROL] = 0x000f81c1;
	}
	return(0);
}


//*****************************************************************************
// PCI-ADADIO - SHUTDOWN
//	This function frees allocated resources.
// Args:
//	iCardCnt	Number of analog I/O cards installed.
//	sCards		This is a pointer to the first element of an
//			array of data structures.  Each array element is a
//			data structure that corresponds to a analog I/O
//			card.  The caller need not manipulate this
//			structure.
// Returns:	
//	1 on success
//	0 on error	
//*****************************************************************************
int iogsadadShutdown(int iCardCnt, AIOCARD *sCards) {

	int iCardNum;

	// Make sure memory has been allocated.
	if (sCards == NULL)
	{
		fprintf(stderr, "%s: NULL pointer to I/O Cards data"
		"array passed to iogsadadShutdown().\n", myname);
		return(1);
	}

	// For each PCI-ADADIO card found...
	for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++)
	{
		// Unmap each PCI-ADADIO register space.
		if (munmap(sCards[iCardNum].pMappedPage, 
		sCards[iCardNum].tMemSize) != 0)
		{
			fprintf(stderr, "%s: Error unmapping PCI-ADADIO "
			"registers via munmap(): %s\n", myname, 
			strerror(errno));
        	        return(1);
		}
	}
	// Return allocated resources here.
	free(sCards);
	sCards = NULL;
	return(0);
}

//*****************************************************************************
// PCI-ADADIO - TRIGGER SAMPLE 
//	This function triggers the analog inputs to collect a frame of sample
//	data.  Where multiple analog boards are installed in a system, they
//	should be configured for hardware triggering off the first module.
//	In other words this function triggers the first analog input module.
//	Any additional analog modules should be automatically triggered via
//	hardware.
// Args:
//	sCards		This is a pointer to the first element of an
//			array of data structures.  Each array element is a
//			data structure that corresponds to a PCI-ADADIO
//			card.  The caller need not manipulate this
// Returns:	
//	0	
//*****************************************************************************
int iogsadadTrig(AIOCARD *sCards)
{
	// Set input trigger in BCR register of the first PCI-ADADIO card.
	sCards[0].pRegisters[BOARD_CONTROL] |= 0x40000000;
	return(0);
}

//*****************************************************************************
// PCI-ADADIO - Copy buffer of ADC data from board to buffer
//	This function collects one frame of data from the analog inputs that 
//	where previously triggered and sampled.  The data is copied to local
//	buffers for further processing.
//	hardware.
// Args:
//	iCardCnt	Number of analog I/O cards installed.
//	sCards		This is a pointer to the first element of an
//			array of data structures.  Each array element is a
//			data structure that corresponds to a PCI-ADADIO
//			card.  The caller need not manipulate this
// Returns:	
//	0 on success
//	-1 on conversion error	
//*****************************************************************************
int iogsadadIn(int InputIndex, int iCardCnt, AIOCARD *sCards)
{
	static	bool	FirstTime = TRUE;
	int	BoardChannel;
	int	ADCchannel;
	word16	Value;
	struct	ioim_inelement *pISRelement;
	int	ConversionError;
	int iCardNum;
	unsigned long ulADCvalue;

	ConversionError = 0;

	// Make sure conversion is complete on each module, if not return error
	// code
	if (!FirstTime)
	{
		for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++)
		{
			// If BUFFER EMPTY bit is set, error out.
			if ((sCards[0].pRegisters[BOARD_CONTROL] &
			0x00100000) != 0)
			{
				ConversionError = -1;
			}
		}
	}

	if (!ConversionError && !FirstTime)
	{

		for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++)
		{
			pISRelement = &(ioim->Queue.Input.Element[InputIndex]);

			// Read each ADC channel into buffers
			ADCchannel = iCardNum * 8;
			for (BoardChannel=0; BoardChannel < 8; BoardChannel++)
			{

				// Read an ADC value
				ulADCvalue = sCards[iCardNum].pRegisters[ANALOG_INPUT_DATA]; 
				Value = ulADCvalue & 0xffff;

				// put a copy into the ISR queue
				pISRelement->Ain[ADCchannel].Value = Value;

				// and also the ADC channel's current value
				// variable
				adc->ioadc[ADCchannel].CrntADCS	= Value;

				// move on to next physical ADC channel
				ADCchannel++;

			} // End for each ADC channel
		} // End for each board
	} // End if

	FirstTime = FALSE;

	return (ConversionError); // Return OK status
}


//*****************************************************************************
// PCI-ADADIO - Output list of values to DAC channels
//	For each DAC channel, this function reads gathers the amplitude and
//	offset from the ISR queue, the ACE gain and offset from the ACE queue,
//	and the Master and Ramp gains from the DAC channel headers.  These 
//	values are used to generate the actual DAC data point and is written
//	to the appropriate DAC channel.
// Args:
//	iCardCnt	Number of analog I/O cards installed.
//	sCards		This is a pointer to the first element of an
//			array of data structures.  Each array element is a
//			data structure that corresponds to a PCI-ADADIO
//			card.  The caller need not manipulate this
// Returns:	
//	0 on success
//	-1 on conversion error	
//*****************************************************************************
int iogsadadOut(int OutputIndex, int iCardCnt, AIOCARD *sCards)
{
	int iCardNum;
			int32			DACchannel;
			int32			ActualDACchannel;
			uint32			DataPoint;
	struct	ioim_aout	  * pDACchannel;
			int16			DataAmplitude;
			TSC_OFFSET		DataOffset;
			TSC_GAIN		ACEgain;
			TSC_OFFSET		ACEoffset;
			TSC_OFFSET		ACEgaincomp;
	struct	ACEchannel    * pACEchannel;
			uint32			ACEindex;
	struct	ioim_ACEparms *	pACEparms;
			TSC_GAIN		MasterGain;
			TSC_GAIN		RampDownGain;
	struct	ioim_dlyd_dp  *	pDelayed;

	// local I/O board register pointers
	volatile uint32 *pDACregister;


	//DEBUG CODE
	//volatile uint32 *pDIOregister;
	//
	//pDIOregister = &(sCards[0].pRegisters[DIGITAL_IO_PORT]);
	//*pDIOregister = 0x4fe; 

	// Wait for any previous simultaneous update to complete. 
	while((sCards[0].pRegisters[BOARD_CONTROL] & 0x20000000)>0); 

	for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++) {

		pDACregister = &(sCards[iCardNum].pRegisters[ANALOG_OUTPUT_CHAN_00]);

		for (DACchannel = 0; DACchannel < 4; DACchannel++) {

			ActualDACchannel	= (iCardNum * 4) + DACchannel;

			pDACchannel			= &(ioim->Queue.Output.Element[OutputIndex].Aout[ActualDACchannel]);
			DataAmplitude		= pDACchannel->Amplitude;
			DataOffset			= pDACchannel->Offset;
			MasterGain			= ioim->cfg.io.Analog.Out.Channel[ActualDACchannel].MasterGain;
			RampDownGain		= pDACchannel->RampGain;

			pACEchannel			= &(ioim->Queue.ACE.DACchannel[ActualDACchannel]);
			ACEindex			= pACEchannel->Head;

			pACEparms			= &(pACEchannel->Element[ACEindex]);
			ACEgain				= pACEparms->Gain;
			ACEoffset			= pACEparms->Offset;
			ACEgaincomp			= pACEparms->GainComp;

			// Adjust data using appropriate ACE parameters
			DataPoint = (uint32)BuildDP(DataAmplitude, DataOffset, ACEgain, ACEoffset, ACEgaincomp, MasterGain, RampDownGain);

			// make sure the value doesn't exceed the maximum of the board
			if (DataPoint >= 0x10000) {
				DataPoint = 0xffff;
			}

			// Save a copy of the new DAC data point to shared memory
			dac->iodac[ActualDACchannel].CrntDACS = (uint16)(DataPoint);

			// Write the data point to the current DAC
			*pDACregister = DataPoint;

			// Point to the next DAC register.
			pDACregister++;

			// Initialize pointer shorthand to avoid re-indexing arrays
			pDelayed = &(ioim->Queue.Delayed.Datapoint[ActualDACchannel]);

			// Save Current ACE parameters in delayed structure

			pDelayed->ACEgain		= ACEgain;
			pDelayed->ACEoffset		= ACEoffset;
			pDelayed->ACEgaincomp	= ACEgaincomp;

			// Update ACE Table Indexes based on cycle status
			switch (pDACchannel->Status & ~SM_FLAG_MASK) {

				case SM_ABOUT_TO_START:
				case SM_END_OF_CYCLE:
				case SM_END_OF_SEGMENT:
				case SM_END_OF_FILE:
				case SM_END_OF_PHASE:

					if (pACEchannel->NumElements == 0) {
						// Set queue under-run flag
						ioim->cfg.Error.Type	= AQ_OUT_OF_SYNC;
						ioim->cfg.Error.Channel	= DACchannel;
					}
					else {
						// Decrement count of elements in ACE queue
						pACEchannel->NumElements--;

						// Increment head pointer to next element
						if (pACEchannel->Head == (ACEPARM_ELEMENTS - 1)) {
							pACEchannel->Head = 0;
						}
						else {
							pACEchannel->Head++;
						}
					}

					break;

				case SM_END_OF_TEST:
				case SM_END_OF_FILE_STOP:
				case SM_END_OF_PHASE_STOP:
				case SM_END_OF_MATRIX:

					// Adjust the currently active ACE gain and offset to it's default
					// (matching that which was placed in the last datapoint array as
					// I/O Manager main() ended the test on this station).

					pACEparms->Gain		= GAIN_1X;
					pACEparms->Offset	= 0;
					pACEparms->GainComp	= 0;

					break;

				default:
					break;
			} // end switch
		} // end for DACchannel
	} // end for iCard	

	// Simultaneously update DAC outputs
	for (iCardNum = 0; iCardNum < iCardCnt; iCardNum++) {
		sCards[iCardNum].pRegisters[BOARD_CONTROL] |= 0x20000000; 
	}

	return(0);
}
