/* COM, Incorporated - Jan. 2002 
 *
 * General Standards Corporation PCI-ADADIO analog I/O card
 * simple D/A output test program for QNX 4.25
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <conio.h>
#include <sys/kernel.h>
#include <sys/pci.h>
#include <string.h>
#include <stddef.h>
#include <sys/seginfo.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

#define VENDOR_ID 0x10b5	// Brand specific PCI bus vendor ID code.
#define DEVICE_ID 0x9080	// PCI device specific ID 
#define SUBSYS_ID 0x2370	// Sub-system vendor ID

#define PCIADADIO_DACS_PER_BOARD 4
#define PCIADADIO_ADCS_PER_BOARD 8

// Register offsets from card base address.
enum RegisterOffsetEnum
{
	BOARD_CONTROL,
	DIGITAL_IO_PORT,
	ANALOG_OUTPUT_CHAN_00,
	ANALOG_OUTPUT_CHAN_01,
	ANALOG_OUTPUT_CHAN_02,
	ANALOG_OUTPUT_CHAN_03,
	ANALOG_INPUT_DATA,
	SAMPLE_RATE
};

// PLX-9080 PCI Interface LCR offsets
// Offset from LCR base address to the interrupt
// control/status register.
#define INT_CTRL_STAT 0x4c

volatile unsigned long		*pRegisters; 			// Pointer to memory mapped registers.
volatile unsigned long 		*pDACregister;
volatile unsigned long 		*pDIOregister;

//**********************************************************
int main (void)
{
	int						iFileDesc;				// File descriptor for physical shared memory.
	int						iInterruptNumber;		// PC Interrupt line (IRQ) number.
	int						iLCRBase;				// Base I/O address of the PCI Local Configuration Registers.
	int						iReturnCode;			// Function return codes
	off_t					tPageAddress;			// Address of page to be mapped.
	size_t					tMemSize;				// Size of mapped memory region.
	unsigned				uBusNum;				// PCI bus number for found device
	unsigned				uDevFuncNum;			// PCI device or function ID for found
	unsigned				uDeviceID = DEVICE_ID;	// Brand specific PCI bus device ID code
	unsigned				uPCIIndex;				// PCI bus index (used to find multiple cards of the same type)
	unsigned				uSubsysID;				// Brand specific PCI bus subsystem ID code
	unsigned				uVendorID = VENDOR_ID;	// Brand specific PCI bus vendor ID code
	unsigned long			DACchannel;
	unsigned long			DataPoint;
	unsigned long			ulPhysicalAddress;		// Physical memory address of registers.
	unsigned long			uPCRValue;				// 32 bit PCI control register values
	unsigned long 			ulOffsetInPage;			// Offset within page to registers.
	void					*pMappedPage;			// Pointer to mapped page(s) containing registers. 
	unsigned long			spinCount;

	// Look for a PCI-ADADIO card on the bus.
	uPCIIndex = 0;
	while(1)	
	{
		// Try to find the next device.
		iReturnCode = _CA_PCI_Find_Device(uDeviceID, uVendorID, uPCIIndex, &uBusNum, &uDevFuncNum);
		if (iReturnCode == PCI_DEVICE_NOT_FOUND) 
		{ // No PCI-ADADIO card found.
			fprintf(stderr, "No PCI-ADADIO I/O card located in the system.\n");
			return(1);
		}
		else if (iReturnCode != PCI_SUCCESS)
		{ // Call error of some sort.
			fprintf(stderr, "Error locating PCI-ADADIO I/O card.\n");
			return(1);
		}
		else		
		{ // Else a card was found.
			// Verify that the Subsystem ID matches that of a PCI-ADADIO card.
			uSubsysID = 0;
			iReturnCode = _CA_PCI_Read_Config_Word(uBusNum, uDevFuncNum, 0x2e, 1, (char *)&uSubsysID);
			if (iReturnCode != PCI_SUCCESS)
			{
				fprintf(stderr, "Error reading subsystem ID from a PCI control register 0x2e via _CA_PCI_Read_Config_Word().\n");
				return(1);
			}
			if (uSubsysID == SUBSYS_ID)
			{
				break;
			}
		}
		uPCIIndex++;
	} // End while

	// Found a card on the PCI bus.
	printf("General Standards PCI-ADADIO AIO card successfully identified\n");

	// Now find it's base address in memory.  Offset 18h
	// of the PCI control register contains the base I/O
	// address of the card in bits 31:2 (along with
	// other information which can be masked off).
	uPCRValue = 0;	
	iReturnCode = _CA_PCI_Read_Config_DWord(uBusNum, uDevFuncNum, 0x18, 1, (char *)&uPCRValue);
	if (iReturnCode != PCI_SUCCESS)
	{
		fprintf(stderr, "Error reading base I/O address from PCI control register 0x18 via _CA_PCI_Read_Config_DWord().\n");
		return(1);
	}
	ulPhysicalAddress = uPCRValue & 0xfffffffc;
			
	// Look up the card's interrupt number.
	uPCRValue = 0;	
	iReturnCode = _CA_PCI_Read_Config_Byte(uBusNum, uDevFuncNum, 0x3c, 1, (char *)&uPCRValue);
	if (iReturnCode != PCI_SUCCESS)
	{
		fprintf(stderr, "Error reading interrupt number from PCI control register 0x3c via _CA_PCI_Read_Config_Byte().\n");
		return(1);
	}
	iInterruptNumber = uPCRValue;

	// Look up the card's PCI Local Configuration
	// Registers (LCR) base address.  Refer to the
	// PLX-9080 PCI interface data sheet for more
	// information.
	uPCRValue = 0;	
	iReturnCode = _CA_PCI_Read_Config_DWord(uBusNum, uDevFuncNum, 0x14, 1, (char *)&uPCRValue);
	if (iReturnCode != PCI_SUCCESS)
	{
		fprintf(stderr, "Error reading base I/O address for PCI Local Configuration Register from PCI control register 0x14 via _CA_PCI_Read_Config_DWord().\n");
		return(1);
	}
	iLCRBase = uPCRValue & 0xfffc;

	// open physical memory
	iFileDesc = shm_open ( "Physical", O_RDWR, 0777);
	if ( iFileDesc == -1)
 	{
		fprintf(stderr, "Error opening physical memory to map PCI-ADADIO configuration registers", strerror(errno));
		return(1);
	}

	// Map the register set of each card into this process space.
	// Since mmap() maps a minimum of a page when mapping physical
	// memory, first it must be determined if the 128 byte register
	// space happens to cross a page boundry.  If the space crosses
	// a page boundry then two pages must be mapped.  In this
	// case this is done by adding 128 to the offset within the
	// page and checking to see if the result is greater than
	// 0x1000 (page size).  If the result is greater, map in two
	// pages, otherwise only one page is required to map in the
	// register set.
	tPageAddress = ulPhysicalAddress &	0xfffff000;
	ulOffsetInPage = ulPhysicalAddress & 0xfff;

	tMemSize = 0x1000;
	if ((ulOffsetInPage + 128) >= 0x1000)
	{
		tMemSize = 0x2000;
	}

	pMappedPage = mmap (0, tMemSize, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_SHARED, iFileDesc, tPageAddress);
	if (pMappedPage == (void *) -1)
	{
		fprintf(stderr, "Error mapping PCI-ADADIO registers via mmap(): %s\n", strerror(errno));
		return(1);
	}
	pMappedPage = pMappedPage;
		
	pRegisters = (unsigned long *)	((char *)pMappedPage + ulOffsetInPage);

	// Initialize each card.  Select the following modes:
	// AIMx (analog input mode): mode 1 - Single ended, burst mode 
	// LBCx (loopback channel): 0
	// CAL STATUS FLAG: Cleared
	// OFFSET BINARY: True
	// SIZEx (input buffer size): 8 samples
	// (Reserved): 0
	// LASTx (last input number): 7 (all inputs used)
	// ENABLE OUTPUTS: True
	// ENABLE OUTPUT STROBE: True
	// INPUT BUFFER EMPTY (read only)
	// INPUT BUFFER HALF FULL (read only)
	// INPUT BUFFER FULL (read only)
	// INTERRUPT Ax (interrupt source): disabled
	// INTERRUPT REQUEST FLAG: clear
	// CMx (calibration mode): none
	// OUTPUT STROBE: 0
	// INPUT TRIGGER: 0
	// INITIALIZE: 0
	pRegisters[BOARD_CONTROL] = 0x000f81c1;
	
	// A sample value to write to DACs.
	DataPoint = 0;

	// Pointer to digital I/O port.
	pDIOregister = &(pRegisters[DIGITAL_IO_PORT]);

	// Set digital output bits high.
	*pDIOregister = 0x4ff; 

	//***************************************************

	while (1)
	{
		// Pointer to DAC 0 register.
		pDACregister = &(pRegisters[ANALOG_OUTPUT_CHAN_00]);

		for (DACchannel = 0; DACchannel < 4; DACchannel++) {

			// ramp DAC sample values.
			DataPoint+=100;

			// make sure the value doesn't exceed the maximum of the board
			if (DataPoint >= 0x10000) {
				DataPoint = 0;
			}

			// Write the data point to the current DAC
			*pDACregister = DataPoint;

			// Pulse a digital output for debug/testing
			*pDIOregister = 0x4fe; 
			*pDIOregister = 0x4ff; 

			// Point to the next DAC register.
			pDACregister++;

		} // end for DACchannel

		// Wait for last update to finish
		while((pRegisters[BOARD_CONTROL] & 0x20000000) !=0);

		// Simultaneously update DAC outputs
		pRegisters[BOARD_CONTROL] |= 0x20000000; 

		// Run until user keystroke
//		if (kbhit()) break;

	} // End while(1)

	//***************************************************

	return(0);
}

